// Copyright (C) 2026 Antoine ANCEAU
//
// This file is part of navaltoolbox.
//
// navaltoolbox is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

//! Tanks module.
//!
//! Provides tank geometry and fluid management.

mod dataclasses;
mod tank;

pub use dataclasses::TankState;
pub use tank::{FSMMode, Tank};

/// A thread-safe shared reference to a Tank.
/// This allows multiple components (Vessel, Python wrappers) to point to the same Tank instance.
pub type SharedTank = std::sync::Arc<std::sync::RwLock<Tank>>;
